#include "protocol.h"

#include <iostream>

#include <cassert>
#include <cstdio>

#if defined(HAVE_CONFIG_H)
#include "config.h"
#endif

// for the min and max templates
#if defined (COMP_VC)
#include "minmax.h"
#else 
#include <algorithm> 
#endif

#include "utils/buffer.h"
#include "basic_types.h"

//TODO: alles plattformabhngig ab hier
// man muesste die endianess und das padding von Header
// durch den compiler bercksichtigen...

static const unsigned int MAX_PACKET_LEN = 1024000;
namespace net
{
  using std::max;

  static const uint_32 PROT_VERSION = 1;
  static const int BUFFER_SIZE = 1024;

  Protocol::Protocol()
    : m_sender(0), m_listener(0),
      bufferBegin(new unsigned char[BUFFER_SIZE]),
      bufferPos(bufferBegin),bufferEnd(bufferBegin+BUFFER_SIZE),
	  m_silent_mode(false)
  {
  }

  Protocol::Protocol(IDataListener& dl)
    : m_sender(0), m_listener(&dl),
      bufferBegin(new unsigned char[BUFFER_SIZE]),
      bufferPos(bufferBegin),bufferEnd(bufferBegin+BUFFER_SIZE),
	  m_silent_mode(false)
  {
  }

  Protocol::~Protocol()
  {
    delete[] bufferBegin;
  }

  void Protocol::registerListener(IDataListener& listener)
  {
    m_listener = &listener;
  }

  void Protocol::registerSender(ISender* sender)
  {
    m_sender = sender;
  }

  namespace {
    struct Header
    {
      uint_32 packetLen;
      uint_32 version;
      uint_32 dataLen;
      uint_32 checkSum;
    };
  }

  static uint_32 check_sum(const Header& h)
  {
    return (h.packetLen << 13) + (h.version << 17) + (h.dataLen << 11);
  }

  int Protocol::send(const utils::Buffer& buf)
  {
	  Header header;
	  
	  header.packetLen = sizeof(header) + buf.getLen();
	  header.version = PROT_VERSION;
	  header.dataLen = buf.getLen();
	  header.checkSum = check_sum(header);
	  
	  if (m_sender)
	  {
		  try {
		  int len = m_sender->send(utils::Buffer(reinterpret_cast<unsigned char*>(&header),
			  sizeof(header)) + buf);
		  m_silent_mode = false;
		  return len;
		  }
		  catch (std::exception& e)
		  {
			  if (!m_silent_mode)
			  {
			    std::cerr << "protocol.cpp: send threw " << e.what() << "\n";
			    std::cerr << "silent mode turned on\n";
				m_silent_mode = true;
			  }			  
		  }
		  
	  }
		  return 0;
  }

  void Protocol::read(const unsigned char*& dataPos,
                      const unsigned char* dataEnd,
                      int bytesToRead, int& bytesLeft)
  {
    if (bytesToRead == 0)
      return;

    assert(bytesToRead > 0);

    if (bufferPos+bytesToRead > bufferEnd)
      {
	int currentSize = bufferEnd - bufferBegin;
	int posDistance = bufferPos - bufferBegin;
	int newSize = max((int) (currentSize*1.3 + 1),
                          (int) (bufferPos-bufferBegin+bytesToRead+1));

	unsigned char* newBuffer = new unsigned char[newSize];
	memcpy(newBuffer,bufferBegin,posDistance);

	delete[] bufferBegin;
	bufferBegin = newBuffer;
	bufferPos = bufferBegin + posDistance;
	bufferEnd = bufferBegin + newSize;
      }

    assert(bufferPos+bytesToRead <= bufferEnd);
    assert(dataPos+bytesToRead <= dataEnd);
    memcpy(bufferPos,dataPos,bytesToRead);
    bufferPos += bytesToRead;
    dataPos   += bytesToRead;
    bytesLeft -= bytesToRead;
  }

  static bool header_ok(const Header& header)
  {
    return (header.version == PROT_VERSION &&
	    header.packetLen <= MAX_PACKET_LEN &&
	    header.packetLen == header.dataLen + sizeof(Header) &&
	    header.checkSum == check_sum(header));
  }

  // tries to find a valid header somwhere inside data
  static int try_sync(const unsigned char* data, int data_len)
  {
    for (unsigned int i = 0; i < data_len - sizeof(Header); ++i)
      {
	const Header* test_header = reinterpret_cast<const Header*>(data+i);
	if (header_ok(*test_header))
	  return i;
      }
    return -1;
  }

  void Protocol::dataReceived(const utils::Buffer& buf)
  {
    const unsigned char* dataBegin = buf.getPtr();
    int len = buf.getLen();

    const unsigned char* dataEnd = dataBegin+len;
    const unsigned char* dataPos = dataBegin;

    int bytesLeft = len;

    Header* header = 0;

    while(dataPos < dataEnd) // noch weitere daten da
      {
	//header noch nicht gelesen?
	if (bufferPos < bufferBegin+sizeof(Header))
	  {
	    // header vervollstndigen
	    if (bufferPos-bufferBegin+bytesLeft < (int) sizeof(Header)) 
	      { // kann header nicht fertig lesen
		// soviel wie geht lesen
		read(dataPos,dataEnd,bytesLeft,bytesLeft);
		assert(dataPos == dataEnd); //alles gelesen
	      } 
	    else // header kann ganz gelesen werden
	      {
		// header fertig lesen
		read(dataPos,dataEnd,bufferBegin-bufferPos+sizeof(Header),
		     bytesLeft);
		assert(bufferPos == sizeof(Header)+bufferBegin);
	      }
	  
	  }
	else // nach header
	  {
	    // bufferbegin kann sich bei read aendern!
	    header = reinterpret_cast<Header *>(bufferBegin);

	    // check if header is valid
	    if (!header_ok(*header))
	      {
		fprintf(stderr, "protocol.cpp: out of sync!\n");
		fprintf(stderr, "pl = %i, dl = %i\n",
			header->packetLen,
			header->dataLen);
		fprintf(stderr, "Trying to resync protocol...\n");

		// Syncing will not work if the header is partly in this data packet
		// and partly in the next.
        int syncPoint = try_sync(dataPos, dataEnd - dataPos);
		if (syncPoint == -1)
		  {
		    fprintf(stderr, "Skipping %i bytes\n", dataEnd - dataPos);
		    bufferPos = bufferBegin;
		    return;
		  }
		else
		  {
		    fprintf(stderr, "Skipping %i bytes\n", syncPoint);
		    dataPos   += syncPoint;
		    bytesLeft -= (syncPoint + sizeof(Header));;
		    bufferPos  = bufferBegin + sizeof(Header);
            
		    memcpy(bufferBegin, dataPos, sizeof(Header));
            dataPos += sizeof(Header);
		    
		    header = reinterpret_cast<Header*>(bufferBegin);
		    fprintf(stderr, "pl = %i, dl = %i\n",
			    header->packetLen,
			    header->dataLen);

		    if (bytesLeft == 0)
		      return;
		  }
	      }

	    if (static_cast<uint_32>(bufferPos-bufferBegin+bytesLeft)
		< header->packetLen) 
	      { // Paket nicht ganz da
		// lesen sovielwie geht
		read(dataPos,dataEnd,bytesLeft,bytesLeft);
		assert(dataPos == dataEnd); // alles gelesen
		// bufferbegin kann sich bei read aendern!
		header = reinterpret_cast<Header *>(bufferBegin);
	      }
	    else // ganzes Paket da
	      {
		// ganzes paket lesen
		read(dataPos,dataEnd,header->packetLen-(bufferPos-bufferBegin),
		     bytesLeft);
	      
		// bufferbegin header kann sich bei read aendern!
		header = reinterpret_cast<Header *>(bufferBegin);

		bufferPos = bufferBegin;
				
		// Paket parsen
		if (header->version == PROT_VERSION && m_listener != 0)
		  {
                    try {
                      m_listener->dataReceived(utils::Buffer(bufferBegin
                                                             +sizeof(Header),
                                                             header->packetLen
                                                             -sizeof(Header)));
                    }
                    catch (std::exception& e)
                      {
                        fprintf(stderr, 
                                "protocol.cpp, dataReceived of listener "
                                "throwed: %s\n", e.what()); 
                      }
                    catch(...)
                      {
                        fprintf(stderr, 
                                "protocol.cpp, dataReceived of listener "
                                "throwed: unknown exception\n");
                      }
		  }
		else if (header->version != PROT_VERSION)
		  {
		    fprintf(stderr, "protocol.cpp: ignoring wrong header version %i\n",
				    header->version);
		  }
	      }

	  } 
	
      } // while
  }

  void Protocol::reset()
  {
    bufferPos = bufferBegin;
  }

} //end of namespace
